/* MIT/X Consortium License
 *
 * Copyright (c) 2012, 2013 Boldarev Artem, East Ukranian National University
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

/* Simple dummy Windows service example. It does NOTHING after installation
 * (it can only start/stop), but you can use it as skeleton for your own.
 */

/* Do not boring me. Windows service should be Unicode application */
#ifndef UNICODE
#define UNICODE
#endif

#ifndef _UNICODE
#define _UNICODE
#endif

/* Disable this damn totally unusable warnings for classic C functions */
#ifndef _CRT_SECURE_NO_WARNINGS
#define _CRT_SECURE_NO_WARNINGS
#endif

#include <windows.h>
#include <process.h>
#include <stdio.h>
#include <wchar.h>
#include <locale.h>

#define SVC_NAME L"FooBarService" /* Internal service name */
#define SVC_DISP_NAME L"FOO BAR Service" /* Service name to display */

/* A double null-terminated array of null-separated names of services
 * or load ordering groups that the system must start before this service.*/
#define SVC_DEPENDENCIES L"" /* Our service has no dependencies */
#define DEF_WAIT_HINT (3000)

/* Wait for main thread timeout. After that main thread will be TERMINATED */
#define STOP_WAIT_TIMEOUT (15000)

/* This structure passed to the main service thread; */
typedef struct _ServiceData {
	DWORD argc; /* Arguments count */
	WCHAR **argv; /* Arguments itself */
	HANDLE stop_event; /* Service stop event */
} ServiceData;

/* 
 * Functions used by service internally
 * and called by Service Control Manager
 */
void WINAPI ServiceMain();
void WINAPI ServiceCtrlHandler(DWORD);
BOOL ReportStatusToSCM(DWORD, DWORD, DWORD);
void AddToMessageLog(WORD, DWORD, WCHAR *);
WCHAR *GetLastErrorText(WCHAR *, DWORD);


/* Service management functions */
int InstallService(WCHAR *svc_name, WCHAR *svc_display_name, WCHAR *executable_path, WCHAR *svc_dependencies);
int RemoveService(WCHAR *svc_name);
int RunService(WCHAR *svc_name);
int StopService(WCHAR *svc_name);

/* Functions that do real work */
int InitService(ServiceData *svc_data);
void ShutdownService(void);
void TerminateService(DWORD);

SERVICE_STATUS ssStatus;
SERVICE_STATUS_HANDLE sshStatusHandle;
/*
 * Service Dispatch Table.
 * Here we specify ServiceMain(),
 * that called by SCM when service starts.
 */
SERVICE_TABLE_ENTRY dispatchTable[] = 
{
	{SVC_NAME, (LPSERVICE_MAIN_FUNCTION) ServiceMain},
	{NULL, NULL}
};

HANDLE ev_stop_service = NULL; /* When we are set this event, service must shutdown */
HANDLE service_thread = NULL; /* Thread that must be started to do real work */

/* 
 * Main function. When called without parameters or by SCM must start service.
 * Also can be used to install/unistall and to control service.
 * Normally user never start service program manually.
 */
int wmain(int argc, WCHAR *argv[])
{
	char *lp = NULL;
	int ret;
	int exit_status = 0;
	WCHAR ebuf[MAX_PATH];

	lp = setlocale(LC_ALL, "");
	ret = 0;
	if ( (argc > 1)&& ( (*argv[1] == L'-') || (*argv[1] == L'/') ))
	{
		if ( wcscmp(L"install", argv[1] + 1) == 0)
		{
			ret = InstallService(SVC_NAME, SVC_DISP_NAME, NULL, SVC_DEPENDENCIES);
			if ( !ret )
			{
				wprintf(L"FAILED: %ls\n", GetLastErrorText(ebuf, MAX_PATH));
				exit_status = 1;
			}
			else if ( ret < 0  )
				wprintf(L"ALREADY: %ls\n", GetLastErrorText(ebuf, MAX_PATH));
			else
			{
				wprintf(L"OK: %ls\n", GetLastErrorText(ebuf, MAX_PATH));
				RunService(SVC_NAME);
			}
		}
		else if ( wcscmp(L"remove", argv[1] + 1 ) == 0)
		{
			ret = RemoveService(SVC_NAME);
			if ( !ret )
			{
				wprintf(L"FAILED: %ls\n", GetLastErrorText(ebuf, MAX_PATH));
				exit_status = 1;
			}
			else if ( ret < 0  )
				wprintf(L"ALREADY: %ls\n", GetLastErrorText(ebuf, MAX_PATH));
			else
				wprintf(L"OK: %ls\n", GetLastErrorText(ebuf, MAX_PATH));
		}
		else if ( wcscmp(L"stop", argv[1] + 1 ) == 0)
		{
			ret = StopService(SVC_NAME);
			if ( !ret )
			{
				wprintf(L"FAILED: %ls\n", GetLastErrorText(ebuf, MAX_PATH));
				exit_status = 1;
			}	
			else if ( ret < 0 )
				wprintf(L"ALREADY: %ls\n", GetLastErrorText(ebuf, MAX_PATH));
			else
				wprintf(L"OK: %ls\n", GetLastErrorText(ebuf, MAX_PATH));
		}
		else if ( wcscmp(L"start", argv[1] + 1 ) == 0)
		{
			ret = RunService(SVC_NAME);
			if ( !ret )
			{
				wprintf(L"FAILED: %ls\n", GetLastErrorText(ebuf, MAX_PATH));
				exit_status = 1;
			}
			else if ( ret < 0  )
				wprintf(L"ALREADY: %ls\n", GetLastErrorText(ebuf, MAX_PATH));
			else
				wprintf(L"OK: %ls\n", GetLastErrorText(ebuf, MAX_PATH));

		}
		else
			goto dispatch;
	}
	else
	{
dispatch:
		if ( !RunService(SVC_NAME) )
		{
			wprintf(L"ERROR: %ls\n\tUsage:\n", GetLastErrorText(ebuf, MAX_PATH));
			wprintf(L"-install to install service\n");
			wprintf(L"-remove to remove service\n");
			wprintf(L"-start to start service\n");
			wprintf(L"-stop to stop service\n");

			exit_status = 1;
		}
		else
			wprintf(L"Service started.\n");
	}
	setlocale(LC_ALL, lp);

	return exit_status;
}

/* Main thread. Add your code here.
 * It should not return while service running.
 */
unsigned int WINAPI ServiceThread(ServiceData *svc_data)
{
	DWORD argc;
	WCHAR **argv;
	HANDLE stop_event;

	/* For using with shorter names. */
	argc = svc_data->argc;
	argv = svc_data->argv;
	stop_event = svc_data->stop_event;

	/* Wait for stop event */
	WaitForSingleObject(stop_event, INFINITE);

	return 0;
}

/* Initialize service thread */
int InitService(ServiceData *svc_data)
{
	service_thread = (HANDLE)_beginthreadex(
			NULL, 							/* security attributes */
			0,							/* thread stack size (0 - default) */
			(unsigned (WINAPI *)( void * )) ServiceThread,		/* pointer to service thread function */
			(void *)svc_data,					/* data, that we pass to thread */
			0,							/* thread creation flag */
			NULL							/* thread ID */
			);

	if ( service_thread == NULL)
		return 0;

	return 1;
}

/* Send event that stops our service */
void ShutdownService(void)
{
	SetEvent(ev_stop_service);
}

/* Free resources and warn SCM that service stopped */
void TerminateService(DWORD err_code)
{
	DWORD exit_code;

	if (sshStatusHandle)
		ReportStatusToSCM(SERVICE_STOPPED, err_code, 0);

	/* If service thread is running we will terminate it,
	 * but normally it must end execution itself.
	 */
	if (service_thread)
	{
		GetExitCodeThread(service_thread, &exit_code);
		TerminateThread(service_thread, exit_code);
		CloseHandle(service_thread);
	}

	if (ev_stop_service)
		CloseHandle(ev_stop_service);

	AddToMessageLog(EVENTLOG_INFORMATION_TYPE, 0, L"Service STOPPED.");
}

/*
 * This function called by SCM and initialize service. It also create stop event.
 * It should not return while service running!
 */
void WINAPI ServiceMain(DWORD argc, WCHAR **argv)
{
	ServiceData svc_data;
	sshStatusHandle = RegisterServiceCtrlHandler(SVC_NAME, ServiceCtrlHandler);

	if ( !sshStatusHandle)
		goto cleanup;

	ssStatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
	ssStatus.dwServiceSpecificExitCode = 0;

	if (!ReportStatusToSCM(SERVICE_START_PENDING, NO_ERROR, DEF_WAIT_HINT))
		goto cleanup;

	/* Initialize Terminate event */
	ev_stop_service = CreateEvent(0, TRUE, FALSE, 0);
	if ( !ev_stop_service )
		goto cleanup;

	/* pass some data for our service thread */
	svc_data.argc = argc;
	svc_data.argv = argv;
	svc_data.stop_event = ev_stop_service;
	if (!InitService(&svc_data))
		goto cleanup;

	/* Report that we are up and running */
	if ( !ReportStatusToSCM(SERVICE_RUNNING, NO_ERROR, 0))
		goto cleanup;
	AddToMessageLog(EVENTLOG_INFORMATION_TYPE, 0, L"Service running...");

	/* Wait for stop event */
	WaitForSingleObject(ev_stop_service, INFINITE);
	/* Wait some time for service thread termination */
	if ( WaitForSingleObject(service_thread, STOP_WAIT_TIMEOUT) == WAIT_OBJECT_0 )
	{
		AddToMessageLog(EVENTLOG_INFORMATION_TYPE, 0, L"Service stopping normally...");
		service_thread = NULL;
	}
	else
	{
		AddToMessageLog(EVENTLOG_ERROR_TYPE, GetLastError(), 
				L"Service shutdown timed out. "
				L"Service will be TERMINATED...");
	}
cleanup:
	if (sshStatusHandle)
		TerminateService(GetLastError());

	return; 
}

/* Here we are handle service control codes */
void WINAPI ServiceCtrlHandler(DWORD dwCtrlCode)
{
	/* Handle the requested control code. */
	switch (dwCtrlCode)
	{
		case SERVICE_CONTROL_STOP:
			ReportStatusToSCM(SERVICE_STOP_PENDING, NO_ERROR, 0);
			ShutdownService();
			return;

			/* Update the service status. */
		case SERVICE_CONTROL_INTERROGATE:
			break;

			/* invalid control code */
		default:
			break;
	}
	ReportStatusToSCM(ssStatus.dwCurrentState, NO_ERROR, 0);
}

/*
 * This function simple fills in the fields of the SERVICE_STATUS structure before calling
 * the API SetServiceStatus() function. However, filling in these fields is repetitive,
 * making the function worthwhile.
 */
BOOL ReportStatusToSCM(DWORD dwCurrentState, DWORD dwExitCode, DWORD dwWaitHint)
{
	static DWORD dwCheckPoint = 1;
	BOOL bResult = TRUE;

	if ( dwCurrentState == SERVICE_START_PENDING )
		ssStatus.dwControlsAccepted = 0;
	else
		ssStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP;

	ssStatus.dwCurrentState = dwCurrentState;
	ssStatus.dwWin32ExitCode = dwExitCode;
	ssStatus.dwWaitHint = dwWaitHint;

	if ( (dwCurrentState == SERVICE_RUNNING) || (dwCurrentState == SERVICE_STOPPED) )
		ssStatus.dwCheckPoint = 0;
	else
		ssStatus.dwCheckPoint = dwCheckPoint++;

	bResult = SetServiceStatus(sshStatusHandle, &ssStatus);
	if ( !bResult )
	{
		AddToMessageLog(EVENTLOG_ERROR_TYPE, GetLastError(), L"SetServiceStatus() error.");
	}
	return bResult;
}

/*
 * Add some message to event log
 */
void AddToMessageLog(WORD wType, DWORD dwErr, WCHAR *lpszMsg)
{
	WCHAR szMsg [(sizeof(SVC_NAME) / sizeof(WCHAR)) + 100 ];
	HANDLE  hEventSource;
	WCHAR *lpszStrings[2];

	/*dwErr = GetLastError(); */

	/* Use event logging to log the error. */
	hEventSource = RegisterEventSource(NULL, SVC_NAME);

	_snwprintf(szMsg, (sizeof(SVC_NAME) / sizeof(WCHAR)) + 100, L"%ls error: %d", SVC_NAME, dwErr);
	lpszStrings[0] = szMsg;
	lpszStrings[1] = lpszMsg;

	if (hEventSource != NULL)
	{
		ReportEvent(hEventSource, /* handle of event source */
				wType,  /* event type */
				0,                    /* event category */
				0,                    /* event ID */
				NULL,                 /* current user's SID */
				2,                    /* strings in lpszStrings */
				0,                    /* no bytes of raw data */
				(const WCHAR **)lpszStrings,          /* array of error strings */
				(LPVOID)NULL);                /* no raw data */

		(void) DeregisterEventSource(hEventSource);
	}
}

/* Translate system error code to text message */
WCHAR *GetLastErrorText( WCHAR *lpszBuf, DWORD dwSize )
{
	DWORD dwRet;
	WCHAR *lpszTemp = NULL;

	dwRet = FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM |FORMAT_MESSAGE_ARGUMENT_ARRAY,
			NULL,
			GetLastError(),
			LANG_NEUTRAL,
			(LPWSTR)&lpszTemp,
			0,
			NULL );

	/* supplied buffer is not long enough */
	if ( !dwRet || ( (long)dwSize < (long)dwRet+14 ) )
		lpszBuf[0] = L'\0';
	else
	{
		if (NULL != lpszTemp)
		{
			lpszTemp[lstrlen(lpszTemp)-2] = L'\0';  /*remove carriage return and newline character */
			swprintf( lpszBuf, dwSize, L"%ls (0x%x)", lpszTemp, GetLastError() );
		}
	}

	if ( NULL != lpszTemp )
		LocalFree((HLOCAL) lpszTemp );

	return lpszBuf;
}

/* Install (register) service.
 * Parameters:
 * svc_name - name of service;
 * svc_display_name - name of service to display;
 * executable_path - path to executable. Can be NULL if you want use current (running) executable file.
 * 	NOTE: This function does not copy executable to the specified path. It must be already there.
 * svc_dependencies - dependencies of service. A pointer to a double null-terminated array of null-separated
 * 	names of services or load ordering groups that the system must start before this service. Can be empty string
 * 	or NULL if service has no dependencies;
 *
 * Return values:
 * >0 - OK;
 * 0 - error;
 * <0 - service already installed.
 *
 * You can use GetLastError() to get detailed error information. (Or It's wrapper GetLastErrorText(), implemented here.)
 *
 * HINT: See help on CreateService() to get more information.
 */
int InstallService(WCHAR *svc_name, WCHAR *svc_display_name, WCHAR *executable_path, WCHAR *svc_dependencies)
{
	SC_HANDLE   svc = NULL;
	SC_HANDLE   svc_manager = NULL;
	int ret;
	WCHAR module_path[MAX_PATH];
	WCHAR *exe_path = NULL;

	/* If executable path not specified, use current executable path */
	if (executable_path == NULL)
	{
		if ( GetModuleFileName( NULL, module_path, MAX_PATH ) == 0 )
			return 0;
		else
			exe_path = module_path;
	}
	else
		exe_path = executable_path;

	ret = 0;
	svc_manager = OpenSCManager(
			NULL,                   /* machine (NULL == local) */
			NULL,                   /* database (NULL == default) */
			SC_MANAGER_CONNECT | SC_MANAGER_CREATE_SERVICE  /* access required */
			);

	if (svc_manager == NULL)
		goto cleanup;

	svc = CreateService(
			svc_manager,			/* SCManager database */
			svc_name,			/* name of service */
			svc_display_name,		/* name to display */
			SERVICE_QUERY_STATUS,		/* desired access */
			SERVICE_WIN32_OWN_PROCESS,	/* service type */
			SERVICE_AUTO_START,		/* start type */
			SERVICE_ERROR_NORMAL,		/* error control type */
			exe_path,			/* service's binary */
			NULL,				/* no load ordering group */
			NULL,				/* no tag identifier */
			svc_dependencies,		/* dependencies */
			NULL,				/* LocalSystem account */
			NULL);				/* no password */

	if ( svc == NULL )
	{
		if ( GetLastError() == ERROR_SERVICE_EXISTS )
			ret = -1;
		goto cleanup;
	}

	ret = 1;

cleanup:
	if (svc_manager)
		CloseServiceHandle(svc_manager);
	if (svc)
		CloseServiceHandle(svc);

	return ret;
}

/* Remove (unregister service) from system.
 * Parameters:
 * svc_name - name of service;
 *
 * Return values:
 * >0 - OK;
 * 0 - error;
 * <0 - service not installed.
 *
 * You can use GetLastError() to get detailed error information. (Or It's wrapper GetLastErrorText(), implemented here.)
 */
int RemoveService(WCHAR *svc_name)
{
	SC_HANDLE   svc = NULL;
	SC_HANDLE   svc_manager = NULL;
	int ret;

	ret = 0;
	svc_manager = OpenSCManager(
			NULL,                   /* machine (NULL == local) */
			NULL,                   /* database (NULL == default) */
			SC_MANAGER_CONNECT   /* access required */
			);

	if (svc_manager == NULL)
		goto cleanup;

	svc = OpenService(svc_manager, svc_name, DELETE | SERVICE_STOP | SERVICE_QUERY_STATUS);
	if (svc == NULL)
	{
		if ( GetLastError() == ERROR_SERVICE_DOES_NOT_EXIST )
			ret = -1;

		goto cleanup;
	}

	/* try to stop service */
	if ( !StopService(svc_name) )
	{
		if ( GetLastError() == ERROR_SERVICE_DOES_NOT_EXIST )
			ret = -1;

		goto cleanup;
	}
	/* now remove the service */
	if ( !DeleteService(svc) )
	{
		if ( GetLastError() == ERROR_SERVICE_DOES_NOT_EXIST )
			ret = -1;
	}
	else
		ret = 1;

cleanup:
	if (svc_manager)
		CloseServiceHandle(svc_manager);
	if (svc)
		CloseServiceHandle(svc);

	return ret;
}

/* Stop running service.
 * Parameters:
 * svc_name - name of service;
 *
 * Return values:
 * >0 - OK;
 * 0 - error;
 * <0 - service not running.
 *
 * You can use GetLastError() to get detailed error information. (Or It's wrapper GetLastErrorText(), implemented here.)
 */
int StopService(WCHAR *svc_name)
{
	SC_HANDLE   svc = NULL;
	SC_HANDLE   svc_manager = NULL;
	SERVICE_STATUS status;
	int ret;
	ret = 0;

	ret = 0;
	svc_manager = OpenSCManager(
			NULL,                   /* machine (NULL == local) */
			NULL,                   /* database (NULL == default) */
			SC_MANAGER_CONNECT   /* access required */
			);

	if (svc_manager == NULL)
		goto cleanup;

	svc = OpenService(svc_manager, svc_name, SERVICE_STOP | SERVICE_QUERY_STATUS);
	if (svc == NULL)
	{
		goto cleanup;
	}

	if ( ControlService( svc, SERVICE_CONTROL_STOP, &status ) == 0 )
	{
		if ( GetLastError() == ERROR_SERVICE_NOT_ACTIVE )
			ret = -1;

		goto cleanup;
	}

	Sleep(250);
	while ( QueryServiceStatus(svc, &status) )
	{
		if ( status.dwCurrentState != SERVICE_STOP_PENDING )
			break;
	}

	ret = status.dwCurrentState == SERVICE_STOPPED ? 1 : 0;

cleanup:
	if (svc_manager)
		CloseServiceHandle(svc_manager);
	if (svc)
		CloseServiceHandle(svc);

	return ret;
}

/* Start installed service.
 * Parameters:
 * svc_name - name of service;
 *
 * Return values:
 * >0 - OK;
 * 0 - error;
 * <0 - service already running.
 *
 * You can use GetLastError() to get detailed error information. (Or It's wrapper GetLastErrorText(), implemented here.)
 */
int RunService(WCHAR *svc_name)
{
	SC_HANDLE   svc = NULL;
	SC_HANDLE   svc_manager = NULL;
	SERVICE_STATUS status;
	int ret;

	/* Should be OK when executable started by Service Control Manager */
	if ( StartServiceCtrlDispatcher(dispatchTable) > 0)
		return 1;

	ret = 0;
	svc_manager = OpenSCManager(
			NULL,                   /* machine (NULL == local) */
			NULL,                   /* database (NULL == default) */
			SC_MANAGER_CONNECT   /* access required */
			);

	if (svc_manager == NULL)
		goto cleanup;

	svc = OpenService(svc_manager, svc_name, SERVICE_START | SERVICE_QUERY_STATUS);
	if (svc == NULL)
		goto cleanup;

	if ( StartService(svc, 0, NULL) == 0)
	{
		if ( GetLastError() == ERROR_SERVICE_ALREADY_RUNNING )
			ret = -1;
		goto cleanup;
	}

	Sleep(250);
	while ( QueryServiceStatus(svc, &status) )
	{
		if ( status.dwCurrentState != SERVICE_START_PENDING )
			break;
	}

	ret = status.dwCurrentState == SERVICE_RUNNING ? 1 : 0;

cleanup:
	if (svc_manager)
		CloseServiceHandle(svc_manager);
	if (svc)
		CloseServiceHandle(svc);

	return ret;
}

